#===========================================================================

package Sitescooper::StoryURLProcessor;

use Sitescooper::URLProcessor;
use Carp;

use strict;

use vars qw(
		@ISA 
	);

@ISA = qw(Sitescooper::URLProcessor);

sub new {
  my $class = shift; $class = ref($class) || $class;
  my ($scoop, $robot, $scf, $ref, $url, $is_dynamic_html, $upindex) = @_;
  my $self = $class->SUPER::new($scoop, $robot, $scf, $ref, $url);

  $self->{is_dynamic_html} = $is_dynamic_html;
  $self->{upindex} = $upindex;

  $self->{origpage} = undef;
  $self->{output_storyurl_dbg} = { };

  # bless ($self, $class);
  $self;
}

sub start_get {
  my $self = shift;
  my $url = $self->{url};
  my $is_dynamic_html = $self->{is_dynamic_html};
  my $upindex = $self->{upindex};

  if ($self->get_state() != $Sitescooper::URLProcessor::STATE_PRE_GET) {
    croak ("state != STATE_PRE_GET");
  }

  my $fullurl = $url; $url = &Scoop::URLWithoutAnchor ($url);
  study $url;

  return if (defined $self->{already_tried_download}{$url});
  $self->{already_tried_download}{$url} = 1;

  my $cacheflag = $self->{scf}->get_story_param ('story_cacheable', $url);
  if (defined $cacheflag) {
    # user setting overrides our heuristics
    $is_dynamic_html = ($cacheflag==0);
  }
  if (defined $self->{scf}->{story_diff} && $self->{scf}->{story_diff}) {
    $is_dynamic_html = 1;	# diff pages are always dynamic
  }

  my $limitto = $self->{scf}->get_story_param ('story_limit_to', $url);
  if (!defined $limitto) { $limitto = $self->{scf}->{def_story_limit_to}; }

  if (defined $limitto) {
    if (!$self->match_url ($url, $limitto)) {
      if (!defined $self->{output_storyurl_dbg}{$self->{url}}) {
	&Scoop::dbg ("StoryURL for $self->{url}: $limitto");
	$self->{output_storyurl_dbg}{$self->{url}} = 1;
      }
      &Scoop::dbg ("Non-story URL ignored: $fullurl");
      return;
    }
  }

  if ($url =~ m,^(ftp|mailto|https|gopher|pnm)://,) {
    &Scoop::dbg ("Non-story URL ignored (bad protocol): $fullurl");
    return;
  }

  my $newurl = $self->apply_url_preproc($url);

  if (!defined $newurl) {
    &Scoop::dbg ("URLProcess says URL should be ignored: $fullurl"); return;
  } elsif ($newurl ne $url) {
    $fullurl = $newurl; $url = &Scoop::URLWithoutAnchor ($newurl);
    $self->{url} = $url;
  }

  my $pat = $self->{scf}->get_story_param ('story_skip', $url);
  if (defined $pat) {
    if ($url =~ m#^${pat}$#) {
      &Scoop::verbose ("Skipping: $fullurl"); return;
    }
  }

  $self->{robot}->check_for_oldest ($url);

  if (!$is_dynamic_html && $Scoop::already_seen {$url}) {
    &Scoop::dbg ("skipping, already seen: $fullurl");
    return;
  }

  push (@Scoop::seen_this_time, $url);
  $Scoop::already_seen {$url} = 1;

  if ($self->{robot}->{hit_file_size_limit}) { return; }
  $self->get_story_page ($url, $is_dynamic_html, $upindex);
  1;
}

# ---------------------------------------------------------------------------

sub get_story_page {
  my $self = shift;
  my $url = shift;
  my $is_dynamic_html = shift;
  my $upindex = shift;

  &Scoop::verbose ("Reading: $url");
  $self->{robot}->check_for_oldest ($url);	# we came across the link, so keep it around

  my $cachedpage = undef;
  my $lastmod;
  if (defined $self->{scf}->{story_diff} && $self->{scf}->{story_diff}) {
    ($cachedpage, $lastmod) = $self->{robot}->get_cached_page_for_diff ($url);
  } elsif ($is_dynamic_html) {
    ($cachedpage, $lastmod) = $self->{robot}->get_cached_page ($url);
  }
  if (defined $cachedpage) {
    $cachedpage = $self->strip_story ($url, $cachedpage, " (cached)");
  }

  return unless ($self->get_page ($url, $is_dynamic_html));

  $self->{cached_story_page} = $cachedpage;
  $self->{is_dynamic_html} = $is_dynamic_html;
  1;
}

# ---------------------------------------------------------------------------

sub finish_get {
  my $self = shift;

  $self->set_state ($Sitescooper::URLProcessor::STATE_POST_GET);
  if ($Scoop::got_intr_flag) { return; }

  my $url = $self->{url};
  my $is_dynamic_html = $self->{is_dynamic_html};
  my $upindex = $self->{upindex};

  my $cachedpage = $self->{cached_story_page};
  delete $self->{cached_story_page};

  my $origpage = $self->get_url_reply();

  if ($self->does_need_http_retry()) {
    my $handler = new Sitescooper::StoryURLProcessor ($self->{scoop},
        $self->{robot}, $self->{scf}, $self->{url},
	$self->get_http_retry_url(), $is_dynamic_html, $upindex);
    $self->{robot}->re_request_page ($handler);
    return;
  }

  return unless defined $origpage;

  if ($Scoop::got_intr_flag) { return; }
  if ($CF::linkslimit > 0 && $self->{robot}->{stories_found} >= $CF::linkslimit) {
    &Scoop::verbose ("over links limit, stopping this site.");
    $self->{robot}->{hit_file_size_limit} = 1;
    return;
  }

  my ($redir_from, $newbase);
  ($redir_from, $url, $newbase) = $self->handle_redirects_and_base_url
  					($url, $origpage);

  # get headline before stripping StoryStart and StoryEnd
  my $headline;
  $headline = $self->get_headline ($url, $origpage);

  $self->{scoop}->journal ("url", $url);
  $self->{scoop}->journal ("base_url", $newbase);
  $self->{scoop}->journal ("pre_strip_story", $origpage);
  my $page = $self->strip_story ($url, $origpage, "");
  $self->{scoop}->journal ("post_strip_story", $page);

  my $bits = $self->{scf}->get_story_param ('story_html_header', $url);
  if (defined $bits) { $page = $bits . $page; }
  $bits = $self->{scf}->get_story_param ('story_html_footer', $url);
  if (defined $bits) { $page .= $bits; }

  my $proc = $self->{scf}->get_story_param ('story_preproc', $url);
  if (defined $proc) {
    $_ = $page;
    my $site_level = 1;
    if (!eval $proc."; 1;") {
      $self->sitewarn ("StoryHTMLPreProc failed: $@");
      # and keep the original $page
    } else {
      $page = $_;
      $self->{scoop}->journal ("post_story_preproc", $page);
    }
  }

  if (defined $self->{scf}->{story_diff} && $self->{scf}->{story_diff}) {
    $page = $self->get_new_bits ($cachedpage, $page);
    $self->{robot}->cache_page_later ($url, $redir_from, $origpage);
  } else {
    $self->{robot}->cache_page ($url, $redir_from, $origpage);
  }

  if ($Scoop::got_intr_flag) { return; }

  if (defined fileno Scoop::JOURNAL) {
    # always write a text-mode version for the journal
    $self->{scoop}->journal ("to_text_story",
    	$self->html_to_text ($url, $page, $Scoop::OUT_TEXT));
  }

  # get turn-over links after stripping StoryStart and StoryEnd
  my @turnoverlinks;
  @turnoverlinks = $self->get_story_turnover_links ($newbase, $page);
  $page = $self->html_to_text_warn_about_ext_links
				 ($url, $page, $self->{robot}->{output_style});

  if ($is_dynamic_html && defined $cachedpage && !$CF::refresh) {
    # ensure that the cleaned-up HTML doesn't match the cleaned-up cached
    # HTML. Sometimes the ad banners will be the only things that have
    # changed between retrieves, and html_to_text will have stripped those
    # out.
    $cachedpage = $self->html_to_text
    			($url, $cachedpage, $self->{robot}->{output_style});
    if ($self->text_equals ($page, $cachedpage)) {
      &Scoop::verbose ("Skipping (text has not changed): $url");
      return;
    }
  }

  my $life = $self->{scf}->get_story_param ('story_lifetime', $url);
  if (defined $Scoop::last_modtime{$url} &&
    		$Scoop::last_modtime{$url} < $life * 24 * 60 * 60)
  {
    &Scoop::verbose ("Skipping (story is older than ".$life." days): $url");
    return;
  }

  my $grepmode = $self->{scf}->get_story_param ('grep', $url);
  if ($grepmode) {
    my $ignprof = $self->{scf}->get_story_param ('ignore_profiles', $url);
    if (!defined $ignprof || $ignprof == 0) {
      my $nhscore;
      if (($nhscore = $self->MatchNewsHoundProfiles
				  ($url, $headline, $page)) eq '')
      {
	&Scoop::verbose ("Skipping (page does not match profiles): $url");
	return;
      }

      #$page = $nhscore.$page;		# TODO??
    }
  }

  # ensure there's some alphanumerics in the output text. No alnums means
  # no output. HTML needs to be checked to ensure we don't just pick
  # up tags which will not be displayed. Added kld's check for image-only
  # sites.
  {
    my $gottext = 1;

    if ($self->{scf}->{image_only_site} == 1) {
      &Scoop::dbg ("image-only site, not checking if text is present");

    } elsif ($self->writing_html) {
      if ($page !~ /[A-Za-z0-9"']\s*</
		&& $page !~ />\s*[A-Za-z0-9"']/
		&& $page !~ /^\s*[A-Za-z0-9"']/)
      { $gottext = 0; }

    } else {
      if ($page !~ /[A-Za-z0-9"']/) { $gottext = 0; }
    }

    if ($gottext == 0) {
      &Scoop::verbose ("Skipping (no text to write): $url");
      return;
    }
  }

  $self->{robot}->add_snarfed_link ($url);
  if ($self->{scf}->{levels} < 0) {
    # this is a one-level site: therefore the story should be treated
    # as the "front page". Thx Carsten for this one.
    $self->{robot}->write_as_story (1, $url, $page, $headline, $upindex);
  } else {
    $self->{robot}->write_as_story (0, $url, $page, $headline, $upindex);
  }

  $self->{scoop}->journal ("turnover_links_story", join ("\n", @turnoverlinks));
  if ($#turnoverlinks >= 0) {
    my $link;

    @turnoverlinks = $self->{robot}->absolutify_and_grep_unseen_urls
					 ($newbase, @turnoverlinks);
    for $link (@turnoverlinks) {
      if ($Scoop::got_intr_flag) { return; }
      $link = &Scoop::AbsoluteURL ($newbase, $link);
      $self->{robot}->download_story_page ($url, $link, 0, $upindex);	# right now
    }
  }
}

# ---------------------------------------------------------------------------

sub MatchNewsHoundProfiles {
  my ($self, $url, $headline, $bodyText) = @_;

  $headline ||= '';

  foreach (@{$self->{scoop}->{profiles}}) {
    my ($score);

    #print "Checking story $headline against profile $_->{name}: ";
    $score = Sitescooper::NewsHound::ScoreStory($_, $bodyText);

    #print "Score $score, needed $_->{score}\n";
    if ($score >= $_->{score}) {
      return "Scored $score on profile $_->{name}\n";
    }
  }
  # if we don't match any profile return null string
  return '';
}

# ---------------------------------------------------------------------------
# Strip a story page from StoryStart to StoryEnd.
# In addition, strip out non-story sidebar table items
# and carriage returns (they confuse plenty of regexps later).
#
sub strip_story {
  my $self = shift;
  my $url = shift;
  my $page = shift;
  my $comment = shift;

  if (!defined $page) { return undef; }

  # ok, now strip the headers and footers
  my $pat = $self->{scf}->get_story_param ('story_start', $url);
  if (defined $pat) {
    if ($page =~ /${pat}.*${pat}/) {
      $self->sitewarn ("StoryStart pattern \"$pat\" found multiple times in page $url$comment\n");
    }
    if ($page =~ s#^.*?${pat}##gs) {
      $page =~ s#^[^<]*?>##gs;		# strip superfluous ends of tags
      if (defined fileno Scoop::JOURNAL) { $self->{scoop}->journal ("pre_stripped", $&); }
    } else {
      $self->sitewarn ("StoryStart pattern \"$pat\" not found in page $url$comment\n");
    }
  }

  $pat = $self->{scf}->get_story_param ('story_end', $url);
  if (defined $pat) {
    if ($page =~ /${pat}.*${pat}/) {
      $self->sitewarn ("StoryEnd pattern \"$pat\" found multiple times in page $url$comment\n");
    }
    if ($page =~ s#${pat}.*?$##gs) {
      $page =~ s#<[^>]*?$##gs;		# strip superfluous starts of tags
      if (defined fileno Scoop::JOURNAL) { $self->{scoop}->journal ("post_stripped", $&); }
    } else {
      $self->sitewarn ("StoryEnd pattern \"$pat\" not found in page $url$comment\n");
    }
  }

  # smart_clean_table only operates on table items with size specifications.
  # TODO -- work out table sizes using images if possible.
  #
  my $smarts = $self->{scf}->get_story_param ('story_use_table_smarts', $url);
  if (!defined $smarts || $smarts != 0) {
    my $filter = Sitescooper::StripTablesFilter->new();
    $filter->parse ($page);
    $page = $filter->filtered_html();
  }

  $page =~ s/\r/ /g;	# strip CRs
  $page;
}

# ---------------------------------------------------------------------------

sub get_headline {
  my $self = shift;
  my $url = shift;
  my $page = shift;

  my $headline;

  if (defined ${$self->{robot}->{url_title}}{$url}) {
    $headline = $self->html_to_text ($url,
    		${$self->{robot}->{url_title}}{$url}, $Scoop::OUT_TEXT);
    &Scoop::dbg ("StoryHeadline: (from RDF): $headline");

  } else {
    my $pat = $self->{scf}->get_story_param ('head_pat', $url);
    if (defined $pat) {
      if ($page !~ m#${pat}#m) {
	$self->sitewarn ("StoryHeadline pattern \"$pat\" not found in page $url\n");
      } elsif (defined $1) {
	$headline = $self->html_to_text ($url, $1, $Scoop::OUT_TEXT);
	# &Scoop::dbg ("StoryHeadline: $headline"); # logged later on anyway
      } else {
	$self->sitewarn ("StoryHeadline pattern \"$pat\" contains no brackets!\n");
      }

    } elsif ($page =~ m#<meta name="PCTITLE" content="\s*(.*?)\s*">#mi) {
      # try a fallback: search for PointCast headline tags
      $headline = $self->html_to_text ($url, $1, $Scoop::OUT_TEXT);
      &Scoop::dbg ("StoryHeadline (default, PointCast): $headline");

    } elsif ($page =~ m#<TITLE>\s*(.*?)\s*</TITLE>#mi) {
      # try a fallback: search for TITLE tag
      $headline = $self->html_to_text ($url, $1, $Scoop::OUT_TEXT);
      &Scoop::dbg ("StoryHeadline (default, TITLE tag): $headline");
    }
  }

  $headline;
}

# ---------------------------------------------------------------------------

sub get_story_turnover_links {
  my $self = shift;
  my $url = shift;
  my $page = shift;

  my @turnoverlinks = ();
  my $followlinks = $self->{scf}->get_story_param
  				('story_follow_links', $url);

  while (1) {
    if ($self->{robot}->{hit_file_size_limit}) { last; }

    if (
      $page =~ s/<a\s+[^>]*href\s*=\s*\"([^\">]+)\"[^>]*>(.+?)<\/a>//is
      ||
      $page =~ s/<a\s+[^>]*href\s*=\s*\'([^\'>]+)\'[^>]*>(.+?)<\/a>//is
      ||
      $page =~ s/<a\s+[^>]*href\s*=\s*([^\s>]+)[^>]*>(.+?)<\/a>//is
      )
    {
      my $link = $1;
      my $txt = $2;

      $link =~ s/^(?:\"|\'|%22)*//; $link =~ s/(?:\"|\'|%22)*$//;
      if ($followlinks) {
	push (@turnoverlinks, $link);

      } elsif (!defined $followlinks && $txt =~ m,(more|next|\d+ of \d+|&gt;&gt;),i)
      {
	# only follow More links automatically if StoryFollowLinks is unset

	my $urlguts = '.';
	($url =~ /^http:\/\/\S+\.([^\.\/]+\.[^\.\/]+\/).*$/) and
	    ($urlguts = $1);
	
	if (($txt !~ /[a-z0-9] [a-z0-9]+ [a-z0-9]+ [a-z0-9]/i) # 5 or more words
	    && (length ($txt) < 15)
	    && $link =~ m/$urlguts/)
	{
	  push (@turnoverlinks, $link);
	  $txt =~ s/[\n\r]+/ /g;
	  &Scoop::verbose ("(Following 'next page' link: \"$txt\")");
	}
      }
      next;
    }

    last;		# no more links available
  }

  @turnoverlinks;
}

# ---------------------------------------------------------------------------

1;
